/*  metadata.cpp */

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

#ifdef WIN32
#include <afxmt.h>
#endif

#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef WIN32
#include <unistd.h>
#include <pthread.h>
#endif

#include <BipCpi.h>

#include "common.h"
#include "mtxstruc.h"
#include "metamem.h"
#include "metadata.h"
#include "trace.h"
#include "parsubs.h"

#define MAX_FILES				10
#define EXTRAFILELEN			32				// padding for metadata area

#ifdef _DEBUG
static char Level[]="metadata.cpp V1.0 Debug version ("__DATE__" "__TIME__")";
#else
static char Level[]="metadata.cpp V1.0 Release version ("__DATE__" "__TIME__")";
#endif

int     useCounter=0;			// Counter for least recently used meta data files
int     filesOpened=0;			// files added to cache
int     filesRemoved=0;			// files removed from cache
int     filesMax=0;				// Maximum number of files in cache

FILETBL* fileTable=NULL;		// File table to cache recent files 

static char fileDirName[256];	// Location of metadata files (mtd)

#ifdef WIN32
static CCriticalSection fileTableLock;	// semaphore for file maniplulation
#else
static pthread_mutex_t fileTableLock=PTHREAD_MUTEX_INITIALIZER;
#endif

#ifdef _DEBUG
extern int debug;
extern int traceParser;
extern int traceWrite;
extern int traceModule;
extern int traceMgmt;
#endif

////////////////////////////////////
//
// Routine to find and read the
// contents of a file into a
// specified area.  
//
////////////////////////////////////

void getFileData(const char * fileName, FILETBL * fileArea)

{
	int		length=0;
	char	*ptr;			// temporary pointer
	char	stdfilename[1024];
	META_FILE_HEADER hdr;	// metadata file header section
	FILE	*stdfile;
#ifdef _DEBUG
	char	errmsg[512];
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> getFileData(%s)\n", fileName);
	}
#endif

	// get the drive and directory for the metadata files
	strcpy(stdfilename, fileDirName);
	strcat(stdfilename, fileName);

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("Opening file %s\n", stdfilename);
	}
#endif

	// try and open the meta data file
	stdfile = fopen(stdfilename, "rb");
	if (stdfile != NULL)
	{
		// determine the length of the definitions data and read
		// it into memory
		// get the length of the metadata
		fseek(stdfile, 0L, SEEK_END);
		length = ftell(stdfile) - META_HDR_LEN;

		// read the header section of the metadata file
		fseek(stdfile, 0L, SEEK_SET);
		fread(&hdr, 1, META_HDR_LEN, stdfile);

		// allocate memory for the metadata and read it
		ptr = (char *) malloc(length);
		fileArea->fvar = (VAR_FILE_REC *) ptr;
		fread(ptr, 1, length, stdfile);

		// Calculate the pointers for the names
		fileArea->namePtr = ptr + hdr.Names - sizeof(hdr);
		ptr += hdr.Unames - sizeof(hdr);
		fileArea->UnamePtr = (CciChar *) ptr;

		// get the number of entries
		fileArea->maxvar = hdr.Maxvar;

		// clear the other entries
		fileArea->flags = 0;
		fileArea->inUseCount = 0;
		fileArea->timesUsed =0;

		// close the file
		fclose(stdfile);
	}
	else
	{
#ifdef _DEBUG
		if (traceModule) 
		{
			sprintf(errmsg, "Unable to open file (%s)\n", stdfilename);
			Log(errmsg);
			traceError(__FILE__, __LINE__, errmsg);
		}
#endif
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- getFileData(length = %d) first name = %.50s\n", length + META_HDR_LEN,fileArea->namePtr);
	}
#endif
}

//////////////////////////////////////
//
// Routine to create a default 
// metadata entry, to be used if
// a message is at a level for which
// no metadata is available.
//
//////////////////////////////////////

void createDefaultEntry()

{
	strcpy(fileTable[0].fileName, "default");
	fileTable[0].maxvar = 0;

	/* allocate a small amount of storage for a dummy names table */
	/* since this storage is set to binary zeros, it can serve    */
	/* as a location for all the file table pointers              */
	fileTable[0].namePtr = (char *)malloc(64);
	memset(fileTable[0].namePtr, 0, 64);
	fileTable[0].UnamePtr = (CciChar *) fileTable[0].namePtr;
	fileTable[0].fvar = (VAR_FILE_REC *) fileTable[0].namePtr;
	fileTable[0].rvar = (REPEAT_REC *) fileTable[0].namePtr;
}

//////////////////////////////////////
//
// Routine to release the metadata 
// for the headers and clear the file
// table entries.                 
//
//////////////////////////////////////

void releaseHeaderMetadata(FILETBL * fileArea)

{
#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> releaseHeaderMetadata()\n");
	}
#endif

	// release the storage for the metadata
	if (fileArea->fvar != NULL)
	{
		free(fileArea->fvar);
	}

	// clear the file table area
	memset(fileArea, 0, sizeof(fileArea));

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- releaseHeaderMetadata()\n");
	}
#endif
}

//////////////////////////////////////
//
// Routine to perform meta data file
// initialization.
//
//////////////////////////////////////

void initializeMetadata()

{
	char*			fileVar;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> initializeMetadata()\n");
	}
#endif

	/* Initialize the file table for the meta data cache in memory */
	fileTable = (FILETBL *) malloc(sizeof(FILETBL) * MAX_FILES);
	if (fileTable != NULL)
	{
		memset(fileTable, 0, sizeof(FILETBL) * MAX_FILES);
	}

#ifdef WIN32
	// Initiatize the file table lock for Windows
	InitializeCriticalSection(fileTableLock);
#endif

	/* check for meta file directory name in environment */
	memset(fileDirName, 0, sizeof(fileDirName));
	fileVar = getenv(METADATADIRENV);
	if (fileVar != NULL)
	{
		// Get the directory name for the meta data files
		// from the environmental variable MTDDIRECTORY
		strncpy(fileDirName, fileVar, sizeof(fileDirName) - 2);
	}
	else
	{
		// set a default
		strcpy(fileDirName, METADATA_DIR_DEF);
	}

	// check for a backslash on the end of the name
	if ((strlen(fileDirName) > 0) && 
		(fileDirName[strlen(fileDirName) - 1] != SLASHCHAR))
	{
		// append a trailing back slash character
		fileDirName[strlen(fileDirName)] = SLASHCHAR;
	}

	/* now, set the first entry to be a default  */
	/* that will be used if no metadata is found */
	/* for a particular version.                 */
	createDefaultEntry();
	
#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- initializeMetadata()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to reset the counter areas used to determine the   */
/*  least recently used meta data file.                       */
/*                                                            */
/**************************************************************/

void resetLRU()

{
	int	i;

	for (i=0; i < MAX_FILES; i++)
	{
		fileTable[i].useCount = 0;
	}
}

/**************************************************************/
/*                                                            */
/* Routine to remove an entry from the metadata file table.   */
/*                                                            */
/**************************************************************/

void removeEntry(const int fileno)

{
#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> removeEntry() fileno=%d\n", fileno);
	}
#endif

	// release the storage for the oldest file
	if (fileTable[fileno].fvar != NULL)
	{
		free(fileTable[fileno].fvar);
	}

	// clear the oldest entry in the table
	fileTable[fileno].useCount = 0;
	memset(fileTable[fileno].fileName, DELETED_FILE_BYTE, sizeof(fileTable[fileno].fileName));
	fileTable[fileno].flags = 0;
	fileTable[fileno].fvar = NULL;
	fileTable[fileno].rvar = NULL;
	fileTable[fileno].maxvar = 0;
	fileTable[fileno].maxrpt = 0;
	fileTable[fileno].timesUsed = 0;

	// decrement the number of files counter
	filesMax--;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- removeEntry()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to find an entry in the table.  If no entry is     */
/*  found, then an entry will be released and an attempt      */
/*  made to load the file into memory.  If the file cannot    */
/*  be loaded, an index of -1 will be returned.               */
/*                                                            */
/**************************************************************/

int findEntry(const char * msgType, const int traceSw)

{
	int	i=0;
	int minUse = 0;
	int minUseCount=0;
	char defFileName[255 + 32];	// name of copy book definitions file 
	META_FILE_HEADER hdr;	// metadata file header section
	char errmsg[512];
    FILE *deffile;			// COBOL definitions file
	char *ptr;				// temporary pointer
	int	defLen=0;			// length of definition data 

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> findEntry() msgType=%s\n", msgType);
	}
#endif

	minUseCount = fileTable[0].useCount + 1;

	// see if we already have this file in the cache
	while ((i < MAX_FILES) && (fileTable[i].fileName[0] > 0) &&
			(strcmp(fileTable[i].fileName, msgType) != 0))
	{
		// keep track of the least used file
		if ((fileTable[i].useCount < minUseCount) && (fileTable[i].inUseCount == 0))
		{
			minUse = i;
			minUseCount = fileTable[i].useCount;
		}

		// Always pick the first deleted entry for reuse if one is found
		if ((fileTable[i].fileName[0] == DELETED_FILE_BYTE) && (fileTable[minUse].fileName[0] != DELETED_FILE_BYTE)) 
		{
			minUse = i;
			minUseCount = fileTable[i].useCount;
		}

		// increment the counter
		i++;
	}

#ifdef _DEBUG
	if (traceSw > 0)
	{
		Log("Meta data index entry is %d, minUse %d %d\n", i, minUse, minUseCount);
	}
#endif

	if ((i >= MAX_FILES) || (fileTable[i].fileName[0] == 0))
	{
		// check if we have to reuse an existing entry
		if (i >= MAX_FILES)
		{
			// clear the least recently used file entry
			removeEntry(minUse);

			// set i to point to the first entry
			i = minUse;

			// increment the files removed from cache counter
			filesRemoved++;
		}

		// try to read the file data
		// build the definitions file name based on the 
		// message type identifier
		memset(defFileName, 0, sizeof(defFileName));

		// Get the drive and directory for the meta data files
		strcpy(defFileName, fileDirName);

		// make sure we have a delimiter between the path name and the file name
		if ((strlen(defFileName) > 0) && 
			(defFileName[strlen(defFileName) - 1] != SLASHCHAR))
		{
			strcat(defFileName, SLASHCHARSTR);
		}

		// append the type name as the first part of the file name
		strcat(defFileName, msgType);

		// Finish the file name by appending the meta data file type
		strcat(defFileName, META_FILE_EXT);

#ifdef _DEBUG
		if (traceSw)
		{
			Log("Opening meta data file - %s\n", defFileName);
		}
#endif

		// Try to open the definitions file
		if ((deffile = fopen(defFileName, "rb")) != NULL)
		{
			// determine the length of the definitions data and read
			// it into memory
			// get the length of the metadata
			fseek(deffile, 0L, SEEK_END);
			defLen = ftell(deffile) - META_HDR_LEN;

			// read the header section of the metadata file
			fseek(deffile, 0L, SEEK_SET);
			fread(&hdr, 1, META_HDR_LEN, deffile);

			// allocate memory for the metadata and read it
			fileTable[i].fvar = (VAR_FILE_REC *) malloc(defLen);
			memset(fileTable[i].fvar, 0, defLen);
			fread(fileTable[i].fvar, 1, defLen, deffile);

			// Calculate the pointers for the names and repeating values
			fileTable[i].namePtr = (char *) fileTable[i].fvar + hdr.Names - sizeof(hdr);
			ptr = (char *) fileTable[i].fvar + hdr.Unames - sizeof(hdr);
			fileTable[i].UnamePtr = (CciChar *) ptr;
			ptr = (char *) fileTable[i].fvar + hdr.RepeatArea - sizeof(hdr);
			fileTable[i].rvar = (REPEAT_REC *) ptr;

			// get the number of variables and repeating entries that we read in
			fileTable[i].maxvar = hdr.Maxvar;
			fileTable[i].maxrpt = hdr.RepeatCount;

			// count this as the first use
			fileTable[i].timesUsed = 1;

#ifdef _DEBUG
			if (traceSw > 0)
			{
				Log("Number of bytes read %d\n into 0x%x\n", defLen,fileTable[i].fvar);
				Log("Names at 0x%x, Unicode names at 0x%x, Repeat data at 0x%x\n",
					fileTable[i].namePtr, fileTable[i].UnamePtr, fileTable[i].rvar);
			}
#endif

			// close the file
			fclose(deffile);

			// Increment the file counters
			filesOpened++;
			filesMax++;

			// set the message type as the name name
			memset(fileTable[i].fileName, 0, sizeof(fileTable[i].fileName));
			strncpy(fileTable[i].fileName, msgType, sizeof(fileTable[i].fileName) - 1);
		}
		else
		{
			sprintf(errmsg, "Unable to open meta data file - %s\n", defFileName);

#ifdef _DEBUG
			if (traceSw)
			{
				Log(errmsg);
			}
#endif

			traceError(__FILE__, __LINE__, errmsg);

			// Indicate file not found
			i = -1;
		}
	}

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- findEntry() i=%d\n", i);
	}
#endif

	return i;
}

/**************************************************************/
/*                                                            */
/* Routine to remove a meta data cache element from memory.   */
/*  This is invoked with a special management message, to     */
/*  allow for online updating of a meta data file without     */
/*  having to shut down the broker.                           */
/*                                                            */
/**************************************************************/

void removeMetaDataEntry(const char *msgType)

{
	int i;
	int found=0;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> removeMetaDataEntry() for %s\n", msgType);
	}
#endif

	// get the semaphore, so that only one thread at a time is
	// working on the cache
#ifdef WIN32
	fileTableLock.Lock();
#else
	pthread_mutex_lock(&fileTableLock);
#endif

	i = 0;
	while (i < MAX_FILES)
	{
		if (memcmp(fileTable[i].fileName, msgType, strlen(msgType)) != 0)
		{
			// found a matching entry, now try to delete it
			found = 1;

			if (fileTable[i].inUseCount == 0)
			{
				// not currently in use, we can delete it safely
				memset(fileTable[i].fileName, DELETED_FILE_BYTE, sizeof(fileTable[i].fileName));
				fileTable[i].useCount = 0;

#ifdef _DEBUG
				if (traceMgmt) 
				{
					Log("Message type(%s) removed from cache\n", msgType);
				}
#endif
			}
			else
			{
				// The file is currently being used
				// flag it for later deletion
				fileTable[i].flags = fileTable[i].flags | DELAYED_DELETION;
				fileTable[i].useCount = 0;

#ifdef _DEBUG
				if (traceMgmt) 
				{
					Log("Message type(%s) scheduled for later removal\n", msgType);
				}
#endif
			}
		}

		i++;
	}

	if (found == 0)
	{
		// the entry is not in the cache
#ifdef _DEBUG
		if (traceMgmt) 
		{
			Log("Message type (%s) not found in cache\n", msgType);
		}
#endif
	}

	// release the semaphore
#ifdef WIN32
	fileTableLock.Unlock();
#else
	pthread_mutex_unlock(&fileTableLock);
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- removeMetaDataEntry()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to remove all meta data cache elements from memory.*/
/*  This is invoked with a special management message, to     */
/*  allow for online updating of a meta data file without     */
/*  having to shut down the broker.                           */
/*                                                            */
/**************************************************************/

void removeAllMetaData()

{
	int i;

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> removeAllMetaData()\n");
	}
#endif

	// get the semaphore, so that only one thread at a time is
	// working on the cache
#ifdef WIN32
	fileTableLock.Lock();
#else
	pthread_mutex_lock(&fileTableLock);
#endif

	i = 1;
	while (i < MAX_FILES)
	{

		// found the entry, now try to delete it
		if (fileTable[i].inUseCount == 0)
		{
			// not currently in use, we can delete it safely
			memset(fileTable[i].fileName, DELETED_FILE_BYTE, sizeof(fileTable[i].fileName));
			fileTable[i].useCount = 0;
		}
		else
		{
			// The file is currently being used
			// flag it for later deletion
			fileTable[i].flags = fileTable[i].flags | DELAYED_DELETION;
			fileTable[i].useCount = 0;
		}

		i++;
	}

	// release the semaphore
#ifdef WIN32
	fileTableLock.Unlock();
#else
	pthread_mutex_unlock(&fileTableLock);
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- removeAllMetaData()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to decrment the in use counter in a meta data      */
/*  cache element, and if need be, remove the entry from      */
/*  memory.                                                   */
/*                                                            */
/**************************************************************/

void decrementUseCount(int traceSw, const int fileIdx)

{
#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: -> decrementUseCount() fileIdx(%d)\n", fileIdx);
	}
#endif

	// get the semaphore, so that only one thread at a time is
	// working on the cache
#ifdef WIN32
	fileTableLock.Lock();
#else
	pthread_mutex_lock(&fileTableLock);
#endif

	// check if the inUseCounter is > 0
	if (fileTable[fileIdx].inUseCount > 0)
	{
		fileTable[fileIdx].inUseCount--;

		// check for a delayed deletion request
		if (((fileTable[fileIdx].flags & DELAYED_DELETION) > 0) && (fileTable[fileIdx].inUseCount == 0))
		{
			// Yes, we need to delete this entry now 
			fileTable[fileIdx].flags = 0;
			memset(fileTable[fileIdx].fileName, DELETED_FILE_BYTE, sizeof(fileTable[fileIdx].fileName) - 1);
			fileTable[fileIdx].useCount = 0;

#ifdef _DEBUG
			if (traceSw) 
			{
				Log("Cache file entry deleted for fileIdx(%d)\n", fileIdx);
			}
#endif
		}
	}
	else
	{
		// something is wrong - the counter should be > 0
#ifdef _DEBUG
		if (traceSw) 
		{
			Log("Invalid use count for fileIdx(%d) inUseCount(%d)\n",
				fileIdx, fileTable[fileIdx].inUseCount);
		}
#endif
	}
	
	// release the semaphore
#ifdef WIN32
	fileTableLock.Unlock();
#else
	pthread_mutex_unlock(&fileTableLock);
#endif

#ifdef _DEBUG
	if (traceModule) 
	{
		Log("PLUGIN: <- decrementUseCount()\n");
	}
#endif
}

/**************************************************************/
/*                                                            */
/* Routine to get the meta data from the meta data file.      */
/*  It keeps a cache of meta data files in memory. The        */
/*  message type is assumed to be a string and not more than  */
/*  6 characters in length.                                   */
/*                                                            */
/**************************************************************/

int getMetaData(const int traceSw,
				const char *msgType)

{
	int i;
	const unsigned char* ptr=NULL;
	int	RC=-1;				// default return code

#ifdef _DEBUG
	if (traceModule)
	{
		Log("Getting meta data for message type %s\n", msgType);
	}
#endif

	// get the semaphore, so that only one thread at a time is
	// working on the cache
#ifdef WIN32
	fileTableLock.Lock();
#else
	pthread_mutex_lock(&fileTableLock);
#endif

	// Search for the message type metadata in memory before we 
	// read the file.  First, we must build the full
	// reference name, which is the name of the metadata
	// file.  It consists of the message type.

	// Either find this entry or clear a space for it in the table
	i = findEntry(msgType, traceSw);

	// Did we find the metadata?
	if (i < 0)
	{
		// Didn't find it - Log the error
#ifdef _DEBUG
		if (traceSw)
		{
			Log("Could not find metadata for %s\n", msgType);
		}
#endif

		RC = -1;
		return RC;
	}

	// increment the usage counter
	useCounter++;
	if (useCounter > (INT_MAX - 5))
	{
		resetLRU();
		useCounter = 1; 
	}

	fileTable[i].useCount = useCounter;
	fileTable[i].timesUsed++;
	fileTable[i].inUseCount++;

	// pass back an index to the data in memory
	RC = i;

#ifdef WIN32
	fileTableLock.Unlock();
#else
	pthread_mutex_unlock(&fileTableLock);
#endif

#ifdef _DEBUG
	if (traceModule > 0)
	{
		Log("Returning data index entry %d\n", i);
	}
#endif

	return RC;
}

const char * getMetadataLevel()

{
	return Level;
}
