/*  Buildmts.c */

/*
 * Licensed Materials - Property of IBM
 * 5648-C63
 * (C) Copyright IBM Corp. 2000
 */

/**************************************************************/
/*                                                            */
/* Program Name:  BUILDMTX.C                                  */
/*                                                            */
/*  Program to build metadata files for the Fix parser.       */
/*                                                            */
/*  Written by Jim MacNair - April, 2001                      */
/*                                                            */
/**************************************************************/

/**************************************************************/
/*                                                            */
/*  Current Version 1.00                                      */
/*                                                            */
/**************************************************************/

/**************************************************************/
/* Required include for C functions                           */
/**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "mtxstruc.h"
#include "comsubs.h"

/**************************************************************/
/*                                                            */
/* Defines and global variables                               */
/*                                                            */
/**************************************************************/

#define MAX_LINE_LEN  2048
#define MAX_VARIABLES 4096
#define MAX_REPEATS	  128

#define INT			"INT"
#define FLOAT		"FLOAT"
#define QTY			"QTY"
#define PRICE		"PRICE"
#define PR_OFFSET	"PRICEOFFSET"
#define AMT			"AMT"
#define CHAR		"CHAR"
#define BOOL		"BOOLEAN"
#define STRING		"STRING"
#define MULTISTR	"MULTIPLEVALUESTRING"
#define CURRENCY	"CURRENCY"
#define EXCHANGE	"EXCHANGE"
#define TIMESTAMP	"UTCTIMESTAMP"
#define TIMEONLY	"UTCTIMEONLY"
#define DATEONLY	"UTCDATE"
#define MARKET_DATE	"LOCALMKTDATE"
#define RAW			"DATA"
#define YYYYMM		"MONTH-YEAR"
#define DAY			"DAY-OF-MONTH"

typedef struct {
   char	VarFlags;
   char	VarFlags2;
   char VarDataType;
   char VarResvd;
   int VarNamePtr;
   int VarUnamePtr;
   int VarTypePtr;
   int VarUtypePtr;
   int VarRawlengthPtr;
} VARSTRUCT;

/**************************************************************/
/* Variable table size definitions                            */
/**************************************************************/

#define MAX_VAR_COUNT  4096

static char copyright[] = "(C) Copyright IBM Corp, 2000";
static char Version[]=\
"@(#)BUILDMTS V1.00 - Fix metadata file processing program  - Jim MacNair ("__DATE__")";

static char nameArea[65536];
static int nameLen=0;
static wchar_t UnameArea[65536];
static int UnameLen=0;

/*******************************************************************/
/*                                                                 */
/* Routine to convert a data type to the coded value.              */
/*                                                                 */
/*******************************************************************/

int getDataType(char * dtype)

{
	int		i;
	int		type=-1;
	char	datatype[MAX_LINE_LEN];

	/* translate the input to upper case */
	i = 0;
	while (dtype[i] != 0)
	{
		datatype[i] = toupper(dtype[i]);
		i++;
	}

	datatype[i] = 0;

	if (strcmp(datatype, INT) == 0)
	{
		type = DATA_INTEGER;
	}
	
	if (strcmp(datatype, FLOAT) == 0)
	{
		type = DATA_FLOAT;
	}
	
	if (strcmp(datatype, QTY) == 0)
	{
		type = DATA_QTY;
	}
	
	if (strcmp(datatype, PRICE) == 0)
	{
		type = DATA_PRICE;
	}
	
	if (strcmp(datatype, PR_OFFSET) == 0)
	{
		type = DATA_PR_OFFSET;
	}
	
	if (strcmp(datatype, AMT) == 0)
	{
		type = DATA_AMT;
	}
	
	if (strcmp(datatype, CHAR) == 0)
	{
		type = DATA_CHAR;
	}
	
	if (strcmp(datatype, BOOL) == 0)
	{
		type = DATA_BOOL;
	}
	
	if (strcmp(datatype, STRING) == 0)
	{
		type = DATA_STRING;
	}
	
	if (strcmp(datatype, MULTISTR) == 0)
	{
		type = DATA_MULTI_STR;
	}
	
	if (strcmp(datatype, CURRENCY) == 0)
	{
		type = DATA_CURRENCY;
	}
	
	if (strcmp(datatype, EXCHANGE) == 0)
	{
		type = DATA_EXCHANGE;
	}
	
	if (strcmp(datatype, TIMESTAMP) == 0)
	{
		type = DATA_TIMESTAMP;
	}
	
	if (strcmp(datatype, DATEONLY) == 0)
	{
		type = DATA_DATEONLY;
	}
	
	if (strcmp(datatype, TIMEONLY) == 0)
	{
		type = DATA_TIMEONLY;
	}
	
	if (strcmp(datatype, MARKET_DATE) == 0)
	{
		type = DATA_MARKET_DT;
	}
	
	if (strcmp(datatype, RAW) == 0)
	{
		type = DATA_RAW;
	}
	
	if (strcmp(datatype, YYYYMM) == 0)
	{
		type = DATA_YYYYMM;
	}
	
	if (strcmp(datatype, DAY) == 0)
	{
		type = DATA_DAY;
	}
	
	return type;
}

void trimString(char * dataArea)

{
	while ((strlen(dataArea) > 0) && (dataArea[strlen(dataArea) - 1] <= ' '))
	{
		dataArea[strlen(dataArea) - 1] = 0;
	}
}

/*******************************************************************/
/*                                                                 */
/* Routine to process a variable definition.                       */
/*                                                                 */
/*******************************************************************/

void writeElement(VARSTRUCT* fvar, FILE* mtdfile, FILE* msgfile)

{
	int				items;
	VAR_FILE_REC	varFileData;

	/* Clear the data area */
	memset(&varFileData, 0, sizeof(varFileData));

	/* set the fields in the file */
	varFileData.Nameptr		= fvar->VarNamePtr;
	varFileData.Unameptr	= fvar->VarUnamePtr;
	varFileData.Flags		= fvar->VarFlags;
	varFileData.Flags2		= fvar->VarFlags2;
	varFileData.Typeptr		= fvar->VarTypePtr;
	varFileData.Utypeptr	= fvar->VarUtypePtr;
	varFileData.Datatype	= fvar->VarDataType;
	varFileData.Ulentype	= fvar->VarRawlengthPtr;

	/* write the data to the file */
	items = fwrite(&varFileData, sizeof(varFileData), 1, mtdfile);
	fflush(mtdfile);

	/* tell what we are doing in the message file */
	fprintf(msgfile, "%4.4s %s\n", nameArea + fvar->VarTypePtr, nameArea + fvar->VarNamePtr);
}

/**************************************************************/
/*                                                            */
/* Routine to write the variable table to a metadata file.    */
/*                                                            */
/**************************************************************/

void writeVars(VARSTRUCT* fvar,
			   REPEAT_REC* rvar,
			   char * mtsname,
			   int maxvar,
			   int maxrpt,
			   FILE* msgfile)

{
	int		i;
	META_FILE_HEADER hdr;
	char	*ptr;
    FILE	*mtsfile;			/* output mtx file                    */
	char	mtsFileName[255];	/* Name of the output meta data file  */

	/* Creat the first part of the file name, using the group identifier */
	strcpy(mtsFileName, mtsname);

	/* check for a file extension */
	ptr = strchr(mtsname, '.');
	if (NULL == ptr)
	{
		/* Finish the meta data file name by adding an extension */
		strcat(mtsFileName, ".mtx");
	}

	if ((mtsfile = fopen(mtsFileName, "wb")) == NULL)
	{
		fprintf(msgfile, "Unable to open meta data file - %s\n", mtsFileName);
	}
	else
	{
		fprintf(msgfile, "Writing meta data file %s\n", mtsFileName);

		/* write out the file header first */
		/* Initialized the header */
		memset(&hdr, 0, sizeof(hdr));

		/* set the fields in the header */
		hdr.HdrLength	= sizeof(hdr);
		hdr.Version		= 0;
		hdr.Maxvar		= maxvar;
		hdr.Names		= sizeof(hdr) + (maxvar * sizeof(VAR_FILE_REC));
		hdr.Unames		= hdr.Names + nameLen + 1;
		hdr.RepeatCount = maxrpt;
		hdr.RepeatArea	= hdr.Unames + ((UnameLen + 1) * 2);

		/* write the header to the metadata file */
		fwrite(&hdr, 1, sizeof(hdr), mtsfile);

		/* write out the individual elements in this meta data file */
		i = 0;
		while (i < maxvar)
		{
			writeElement(&fvar[i], mtsfile, msgfile);
			i++;
		}

		/* write out the names and Unicode names areas to the metadata file */
		fwrite(nameArea, 1, nameLen + 1, mtsfile);
		fwrite(UnameArea, 1, 2 * (UnameLen + 1), mtsfile);

		/* write out the repeating items definition table */
		fwrite(rvar, sizeof(REPEAT_REC), maxrpt, mtsfile);

		/* Close the metadata file */
		fclose(mtsfile);
	}
}

char * processTag(char * ptr)

{
	int		size;
	char	*endptr;
	wchar_t	UtempType[32];

	assert(ptr != NULL);

	endptr = strchr(ptr, ',');
	if (endptr != NULL)
		endptr[0] = 0;

	/* convert the tag to Unicode, for efficiency at runtime */
	size = mbstowcs(UtempType, ptr, sizeof(UtempType) - 1);

	/* append it to the Unicode name table */
	wcscpy(UnameArea + UnameLen, UtempType);
	UnameLen += wcslen(UtempType) + 1;

	/* prepare to handle the next item, if it exists */
	/* otherwise, return a pointer to a NULL string  */
	if (endptr != NULL)
		endptr++;
	else
		endptr = ptr + strlen(ptr);

	return endptr;
}

void saveName(const char * name)

{
	int		size;
	wchar_t	UtempName[256];

	if (strlen(name) > 0)
	{
		/* first, append this name to the name table */
		strcpy(nameArea + nameLen, name);
		nameLen += strlen(name) + 1;

		/* convert the long name to Unicode, for efficiency at runtime */
		size = mbstowcs(UtempName, name, sizeof(UtempName) - 1);

		/* append it to the Unicode name table */
		wcscpy(UnameArea + UnameLen, UtempName);
		UnameLen += wcslen(UtempName) + 1;
	}
}

/**************************************************************/
/*                                                            */
/* Routine to write the input file name to the message file.  */
/*                                                            */
/**************************************************************/

void writeFileName(char *fileName, FILE *msgfile)

{
	/* write the file name to the message file */
	fprintf(msgfile, "Processing file %s\n", fileName);
	fflush(msgfile);
}

/**************************************************************/
/*                                                            */
/* Routine to write the line counter to the message file.     */
/*                                                            */
/**************************************************************/

void writeLinesRead(int inCount, FILE *msgfile, char *filename)

{
	/* write the line count to the message file */
   fprintf(msgfile,"Input lines read %5d  file %s\n",inCount,filename);
}

/**************************************************************/
/*                                                            */
/* Routine to display the command syntax.                     */
/*                                                            */
/**************************************************************/

void printHelp(char *cmd)
{
   printf("\nBUILDMTX V1.00\n");
   printf("\nCommand usage is:\n\n");
   printf("  %s Fix-definitions-file Repeating-definitions-file Fix-metadata-file\n\n", cmd);
   printf("A report file is printed with with the name of\n");
   printf("the fix definitions file and an extension of msg\n");
   printf("The file extension should be omitted from the metadata");
   printf(" file name.  An extension of \"mtx\" will be appended.\n");
}

/**************************************************************/
/*                                                            */
/* Routine to create the meta data files from the ASC data    */
/*  dictionary file.                                          */
/*                                                            */
/**************************************************************/

void main(int argc, char *argv[])

{
	int  maxvar=0;          /* Number of fields found             */
	int  maxrpt=0;			/* Number of repeat table entries     */
	int	 rc=0;				/* Return code from program           */
	int  dataType;			/* Variable type                      */
	int  linecount=0;		/* line number within a data type     */
	int  flags=0;			/* variable flags                     */
	int  flags2=0;			/* variable flags2                    */
	char *ptr;				/* work variable                      */
	char *endptr;			/* work variable                      */
	char *fileExt;          /* file extension                     */
	char dataArea[MAX_LINE_LEN];
	char nameArea[MAX_LINE_LEN];
	char typeArea[MAX_LINE_LEN];
	char dataTypeArea[MAX_LINE_LEN];
	char rawLengthType[MAX_LINE_LEN];
	char fileName[32];		/* Work area to save file name        */
	char msgFileName[255];  /* name of message file               */
	char dataFileName[255]; /* name of input definitions file     */
	char rptFileName[255];  /* name of repeating definitions file */
	char mtxFileName[255];  /* name of output metadata file       */
	VARSTRUCT fvar[MAX_VARIABLES];
	REPEAT_REC rvar[MAX_REPEATS];

 /*
  * File variables and counters
  */

    FILE        *msgfile;		/* conversion program messages */
    FILE        *datafile;		/* input data file             */
	FILE		*repeatfile;	/* repeating definitions file  */
    int         inCounter=0;	/* input line counter          */
    int         outCounter=0;	/* output line counter         */
	int			inFile=0;		/* Lines within current file   */
	int			linesSkipped=0;	/* comment lines              */
	int			badlines=0;		/* Invalid lines               */

/**************************************************************/
/*     Allocate and initialize variable areas                 */
/**************************************************************/

	memset(fileName, 0, sizeof(fileName));
	memset(fvar, 0, MAX_VARIABLES * sizeof(VARSTRUCT));
	memset(rvar, 0, MAX_REPEATS * sizeof(REPEAT_REC));
	memset(nameArea, 0, sizeof(nameArea));
	nameLen = 1;
	memset(UnameArea, 0, sizeof(UnameArea));
	UnameLen = 1;

/**************************************************************/
/*     Parse the file name and generate new file name.        */
/**************************************************************/

	if (argc >= 4)        /* check for file names */
	{
		/* Get the file name parameter */
		strcpy(dataFileName,argv[1]);

		/* check for help request */
		if ((dataFileName[0] == '?') || (dataFileName[1] == '?'))
		{
			printHelp(argv[0]);
			exit(1);
		} /* endif */
	}
	else
	{
		/* must have two file names, one for copy book and one for data file */
		printHelp(argv[0]);
		exit(1);
	} /* endif */


/**************************************************************/
/* Generate the new file name                                 */
/**************************************************************/

	/* get the repeating data file name */
	strcpy(rptFileName, argv[2]);

	/* get the metadata file name */
	strcpy(mtxFileName, argv[3]);

	/* get the full file name */
	strcpy(msgFileName,dataFileName);

	/* Remove any extension from the message file name */
	/* look for a period */
	fileExt = strchr(msgFileName, '.');
	if (fileExt != NULL)
	{
		/* remove the file extension if it exists */
		(*fileExt) = 0;
	} /* endif */

   /* add a file extension to the message files */
   strcat(msgFileName,".msg");

/**************************************************************/
/* Open the input file, data file and message file.           */
/**************************************************************/

	if ((datafile = fopen(dataFileName, "r")) == NULL)
	{
		printf("Unable to open input file - %s\n", dataFileName);
		exit(100);
	}

	if ((repeatfile = fopen(rptFileName, "r")) == NULL)
	{
		fclose(datafile);
		printf("Unable to open repeating definitions file - %s\n", rptFileName);
		exit(101);
	}

	if ((msgfile = fopen(msgFileName, "w")) == NULL)
	{
		fclose(datafile);
		fclose(repeatfile);
		printf("Unable to open message file - %s\n", msgFileName);
		exit(102);
	}

	/* Write the name of the input file to the message file */
	writeFileName(dataFileName, msgfile);

/**************************************************************/
/*                                                            */
/*  This routine will read the meta data definition file      */
/*  and generate a listing file and a meta data file.         */
/*                                                            */
/**************************************************************/

   /* Read the next line(s) in the source file. */
	while (1)
	{
		/* Get the next line */
		if (fgets(dataArea, MAX_LINE_LEN, datafile) == NULL)
		{
			/* return if end of file is reached */
			break;
		}

		/* remove any trailing newlines, etc */
		trimString(dataArea);

		if ((dataArea[0] != 0) && (dataArea[0] != '*'))
		{
			/* initialize variable areas */
			flags2 = 0;
			rawLengthType[0] = 0;

			/* break the line up into individual parts */
			ptr = dataArea;

			/* isolate and capture the tag */
			endptr = strchr(ptr, ',');
			endptr[0] = 0;
			strcpy(typeArea, ptr);

			/* isolate and capture the name */
			ptr = endptr + 1;
			endptr = strchr(ptr, ',');
			endptr[0] = 0;
			strcpy(nameArea, ptr);

			/* isolate and capture the data type */
			ptr = endptr + 1;
			endptr = strchr(ptr, ',');
			endptr[0] = 0;
			strcpy(dataTypeArea, ptr);

			/* capture the flags value */
			ptr = endptr + 1;
			endptr = strchr(ptr, ',');
			if (endptr != NULL)
			{
				/* terminate the flags entry */
				endptr[0] = 0;
				flags = atoi(ptr);

				/* point to the flags2 entry */
				ptr = endptr + 1;
				endptr = strchr(ptr, ',');
				if (endptr != NULL)
				{
					endptr[0] = 0;
					flags2 = atoi(ptr);
					/* find the raw data length pointer */
					ptr = endptr + 1;
					strcpy(rawLengthType, ptr);
				}
				else
				{
					flags2 = atoi(ptr);
				}
			}
			else
			{
				flags = atoi(ptr);
				flags2 = 0;
			}

			/* convert the data type into an encoded value */
			dataType = getDataType(dataTypeArea);

			/* process this segment */
			fvar[maxvar].VarDataType = dataType;
			fvar[maxvar].VarNamePtr = nameLen;
			fvar[maxvar].VarUnamePtr = UnameLen;
			saveName(nameArea);
			fvar[maxvar].VarTypePtr = nameLen;
			fvar[maxvar].VarUtypePtr = UnameLen;
			saveName(typeArea);
			fvar[maxvar].VarFlags = flags;
			fvar[maxvar].VarFlags2 = flags2;
			fvar[maxvar].VarRawlengthPtr = 0;
			if (rawLengthType[0] != 0)
			{
				fvar[maxvar].VarRawlengthPtr = UnameLen;
				saveName(rawLengthType);
			}

			/* Count the number of segments found */
			maxvar++;
		}
	}

	/* Write the name of the repeating definitions file to the message file */
	writeFileName(dataFileName, msgfile);

	/* process the repeating definitions file */
	while (1)
	{
		/* Get the next line */
		if (fgets(dataArea, MAX_LINE_LEN, repeatfile) == NULL)
		{
			/* return if end of file is reached */
			break;
		}

		/* remove any trailing newlines, etc */
		trimString(dataArea);

		/* check for comments */
		if ((0 != dataArea[0]) && ('*' != dataArea[0]))
		{
			/* break the line into items */
			ptr = dataArea;
			endptr = strchr(ptr, ',');
			endptr[0] = 0;
			strcpy(rvar[maxrpt].Name, ptr);

			/* get the count tag */
			ptr = endptr + 1;

			/* point to where the tag will be stored */
			rvar[maxrpt].CountUtypeptr = UnameLen;
			ptr = processTag(ptr);

			/* get the first item tag */
			rvar[maxrpt].FirstUtypeptr = UnameLen;
			ptr = processTag(ptr);

			/* get the parent element name */
			rvar[maxrpt].Nameptr = UnameLen;
			ptr = processTag(ptr);

			/* finally, we need to capture the extra items that can */
			/* be part of the repeating item */
			rvar[maxrpt].ExtraTypes = UnameLen;
			while ((ptr != NULL) && (ptr[0] != 0))
			{
				ptr = processTag(ptr);
			}

			/* insert a terminating null item */
			UnameArea[UnameLen++] = 0;

			/* move on to the next entry in the table */
			maxrpt++;
		}
	}

	/* Check if we found any variables */
	if (maxvar > 0)
	{
		/* write out the variables into the mtd file */
		writeVars(fvar, rvar, mtxFileName, maxvar, maxrpt, msgfile);

		/* Write the number of lines read to the message file */
		writeLinesRead(maxvar, msgfile, fileName);
	}

	/* write the counters to the message file */
	fprintf(msgfile, "\nTotal lines processed in metadata definitions file %d\n", linecount);
	fprintf(msgfile, "Total comment lines skipped %d\n", linesSkipped);
	fprintf(msgfile, "Total invalid lines skipped %d\n", badlines);
	fprintf(msgfile, "Total repeating item lines processed %d\n", maxrpt);

	/* close the files */
	fclose(datafile);
	fclose(msgfile);
	fclose(repeatfile);

/**************************************************************/
/* Finished - return normally                                 */
/**************************************************************/

	return;
}

/*-------------*/
/* End Of File */
/*-------------*/